unit UnsharpMask;

interface

uses
  Windows, AviUtlPlugin, UtilType;

type
  PFilterParam = ^TFilterParam;
  TFilterParam = record
    WeightTableBuf: Pointer;
    WeightTableInfo: array of DWORD;
    Length: UINT;
    Step: UINT;
    BLoopCount: UINT;
    XLoopCount: UINT;
    YLoopCount: UINT;
  end;

  TGaussianBlurWeight = class
  protected
    FRadius: Double;
    FSigma: Double;
    WeightTable: array of Double;
    FIRadius: Integer;
    function WindowFunc(d: Integer): Double;
    procedure SetWeightTable;
  public
    constructor Create(_Radius, _Sigma: Double);
    procedure SetWeight(PPrm: PFilterParam);
    procedure SetParam(_Radius, _Sigma: Double);
    property Radius: Double read FRadius;
    property Sigma: Double read FSigma;
    property IRadius: Integer read FIRadius;
  end;

  TUnsharpMaskFilter = class
  private
    SrcBpl: UINT;
    TmpBpl: UINT;
    HFilterSrcUnitSize: UINT;
    HFilterDestUnitSize: UINT;
    VFilterSrcUnitSize: UINT;
    VFilterDestUnitSize: UINT;
    HFilterWorkSize: UINT;
    HFilterWork: Pointer;
    BlurWeight: TGaussianBlurWeight;
    MaxThread: UINT;
    Threshold: Integer;
    StrengthPlus: Integer;
    StrengthMinus: Integer;
    HPrm: TFilterParam;
    VPrm: TFilterParam;
  public
    constructor Create(_MaxThread: UINT);
    destructor Destroy; override;
    procedure ResetParam;
    procedure SetFilterParam(Width,Height,MaxWidth: UINT);
    procedure SetBlurParam(_Radius, _Sigma: Double);
    procedure SetStrength(Plus, Minus: Integer);
    procedure SetThreshold(_Threshold: Integer);
    procedure HFilter(Dest, Src: Pointer; ThreadID, ThreadNum: UINT);
    procedure VFilter(Dest, Src: Pointer; ThreadID, ThreadNum: UINT);
  end;

var
  USM: TUnsharpMaskFilter;

{$L asm\unsharpmask.obj}

procedure _luma_hfilter(Dest, Src, WeightTable, Work: Pointer; BLoopCount, XLoopCount, YLoopCount, DestBpl, SrcBpl: UINT); cdecl; external;
procedure _unsharpmask_vfilter(Dest, Src, WeightTable: Pointer; XLoopCount, YLoopCount, DestBpl, SrcBpl: UINT); cdecl; external;
procedure _setthreshold(Threshold: Integer); cdecl; external;
procedure _setstrength(Plus, Minus: Integer); cdecl; external;

implementation

uses
  Math, xmmutil;

{TGaussianBlurWeight}
constructor TGaussianBlurWeight.Create(_Radius, _Sigma: Double);
begin
  inherited Create;
  SetParam(_Radius, Sigma);
end;

procedure TGaussianBlurWeight.SetWeightTable;
var
  i: Integer;
  d: Double;
begin
  SetLength(WeightTable, IRadius+1);
  for i:=0 to IRadius do
  begin
    d := i * 8 / Radius;
    WeightTable[i] := Exp(-d*d/Sigma);
  end;
end;

procedure TGaussianBlurWeight.SetParam(_Radius, _Sigma: Double);
begin
  if (Radius <> _Radius) or (Sigma <> _Sigma) then
  begin
    FRadius := _Radius;
    FSigma := _Sigma;
    FIRadius := Trunc(_Radius);
    if IRadius <> _Radius then Inc(FIRadius);
    SetWeightTable();
  end;
end;

function TGaussianBlurWeight.WindowFunc(d: Integer): Double;
begin
  d := Abs(d);
  if d <= IRadius then
    Result := WeightTable[d]
  else
    Result := 0;
end;

procedure TGaussianBlurWeight.SetWeight(PPrm: PFilterParam);
var
  Tap, TapPair: UINT;
  Work: array of Double;
  p,p2,p3: PIntArray;
  i, j: Integer;
  First, Last: Integer;
  FirstPixel: Integer;
  LastTapIndex: Integer;
  Offset: Integer;
  Sum: Double;
  pWeight: PSmallintArray;
  Weight: Double;
begin
  Tap := IRadius * 2 + 1;
  TapPair := IRadius + 1;
  p := GetAlignMem((IRadius * 2 + 2) * (TapPair * 4 + 8), 8);
  SetLength(Work, TapPair * 2);
  SetLength(PPrm.WeightTableInfo, PPrm.Length*2+2);
  PPrm.WeightTableBuf := p;
  p := @p[1];
  First := -IRadius-1;
  Last := First + Tap - 1;
  for i:=0 to PPrm.Length-1 do
  begin
    Inc(First);
    Inc(Last);
    if  First >= 0 then
      FirstPixel := First
    else
      FirstPixel := 0;
    Offset := FirstPixel and 1;
    if Last < PPrm.Length then
      LastTapIndex := Last - FirstPixel
    else
    begin
      LastTapIndex := PPrm.Length - 1 - FirstPixel;
      Work[LastTapIndex+Offset] := 0;
    end;
    PPrm.WeightTableInfo[i*2] := (FirstPixel shr 1) * PPrm.Step;
    if (i > IRadius+1) and (i < PPrm.Length-IRadius) then
    begin
      if Offset = 0 then PPrm.WeightTableInfo[i*2+1] := DWORD(p2)+4
      else PPrm.WeightTableInfo[i*2+1] := DWORD(p3)+4;
      Continue;
    end;
    p[0] := (LastTapIndex + 1+ Offset + 1) div 2;
    PPrm.WeightTableInfo[i*2+1] := DWORD(p)+4;
    Sum := 0;
    Work[0] := 0;
    Work[(LastTapIndex+OffSet) or 1] := 0;
    for j:=First to Last do
    begin
      Weight := WindowFunc(j-i);
      Sum := Sum + Weight;
      if j > 0 then
      begin
        if j < PPrm.Length then
          Work[j - FirstPixel + Offset] := Weight
        else
        begin
          Work[j - FirstPixel + Offset] := 0;
          Work[LastTapIndex + Offset] := Work[LastTapIndex + Offset] + Weight;
        end;
      end
      else
        Work[0] := Work[0] + Weight;
    end;
    if (i = IRadius) or (i = IRadius+1) then
    begin
      if Offset = 0 then p2 := @p[0]
      else p3 := @p[0];
    end;
    pWeight := @p[1];
    for j:=0 to p[0] * 2 - 1 do
      pWeight[j] := Trunc(Work[j] / Sum * $4000);
    p := Pointer(((DWORD(@pWeight[p[0] * 2 + 2]) +7) and not 7) - 4);
  end;
end;

{TUnsharpMaskFilter}
constructor TUnsharpMaskFilter.Create(_MaxThread: UINT);
begin
  inherited Create();
  MaxThread := _MaxThread;
  BlurWeight := TGaussianBlurWeight.Create(2.5, 2);
end;

destructor TUnsharpMaskFilter.Destroy;
begin
  ResetParam();
  BlurWeight.Free();
  inherited Destroy();
end;

procedure TUnsharpMaskFilter.SetBlurParam(_Radius, _Sigma: Double);
begin
  if (BlurWeight.Radius <> _Radius) or (BlurWeight.Sigma <> _Sigma) then
  begin
    BlurWeight.SetParam(_Radius, _Sigma);
    ResetParam();
  end;
end;

procedure TUnsharpMaskFilter.SetStrength(Plus, Minus: Integer);
begin
  if (StrengthPlus <> Plus) or (StrengthMinus <> Minus) then
  begin
    StrengthPlus := Plus;
    StrengthMinus := Minus;
    _setstrength(-Plus, -Minus);
  end;
end;

procedure TUnsharpMaskFilter.SetThreshold(_Threshold: Integer);
begin
  if Threshold <> _Threshold then
  begin
    Threshold := _Threshold;
    _setthreshold(Threshold);
  end;
end;

procedure TUnsharpMaskFilter.ResetParam;
  procedure ClearParam(PPrm: PFilterParam);
  begin
    if PPrm.WeightTableBuf <> nil then
    begin
      FreeAlignMem(PPrm.WeightTableBuf);
      PPrm.WeightTableBuf := nil;
    end;
    PPrm.WeightTableInfo := nil;
    ZeroMemory(PPrm, SizeOf(TFilterParam));
  end;
begin
  ClearParam(@HPrm);
  ClearParam(@VPrm);
  if HFilterWork <> nil then
  begin
    FreeAlignMem(HFilterWork);
    HFilterWork := nil;
  end;
end;

procedure TUnsharpMaskFilter.SetFilterParam(Width,Height,MaxWidth: UINT);
begin
  if (Width = HPrm.Length) and (Height = VPrm.Length) then Exit;
  ResetParam();
  HPrm.Length := Width;
  VPrm.Length := Height;
  SrcBpl := SizeOf(PIXEL_YC) * MaxWidth;
  TmpBpl := ((Width+7) and not 7) * 2;
  HPrm.Step := SizeOf(SmallInt) * 8;
  HPrm.XLoopCount := Width;
  HPrm.YLoopCount := (Height+3) div 4;
  HPrm.BLoopCount := (Width+3) div 4;

  VPrm.XLoopCount := (Width+7) div 8;
  VPrm.YLoopCount := Height;
  VPrm.Step := TmpBpl * 2;
  HFilterSrcUnitSize := SrcBpl * 4;
  HFilterDestUnitSize := TmpBpl * 4;
  VFilterSrcUnitSize := 32;
  VFilterDestUnitSize := 48;
  HFilterWorkSize := ((((Width + 3) div 4) * HPrm.Step * 2) + HPrm.Step + 127) and not 127;
  if HFilterWorkSize <> 0 then
    HFilterWork := GetAlignMem(HFilterWorkSize * MaxThread, 128);
  BlurWeight.SetWeight(@HPrm);
  BlurWeight.SetWeight(@VPrm);
end;

procedure TUnsharpMaskFilter.HFilter(Dest, Src: Pointer; ThreadID, ThreadNum: UINT);
var
  head, tail: UINT;
  YLoopCount: UINT;
begin
  if ThreadNum > MaxThread then ThreadNum := MaxThread;
  if ThreadID >= ThreadNum then Exit;
  if ThreadNum <> 1 then
  begin
    head := HPrm.YLoopCount * ThreadID div ThreadNum;
    tail := HPrm.YLoopCount * (ThreadID+1) div ThreadNum;
    Src := Pointer(DWORD(Src) + head * HFilterSrcUnitSize);
    Dest := Pointer(DWORD(Dest) + head * HFilterDestUnitSize);
    YLoopCount := tail - head;
  end
  else
    YLoopCount := HPrm.YLoopCount;
  if YLoopCount <> 0 then
    _luma_hfilter(Dest, Src, @HPrm.WeightTableInfo[0], Pointer(UINT(HFilterWork)+ThreadID*HFilterWorkSize), HPrm.BLoopCount, HPrm.XLoopCount, YLoopCount, TmpBpl, SrcBpl);
end;

procedure TUnsharpMaskFilter.VFilter(Dest, Src: Pointer; ThreadID, ThreadNum: UINT);
var
  head, tail: UINT;
  XLoopCount: UINT;
begin
  if ThreadNum > MaxThread then ThreadNum := MaxThread;
  if ThreadID >= ThreadNum then Exit;
  if ThreadNum <> 1 then
  begin
    head := VPrm.XLoopCount * ThreadID div ThreadNum;
    tail := VPrm.XLoopCount * (ThreadID+1) div ThreadNum;
    Src := Pointer(DWORD(Src) + head * VFilterSrcUnitSize);
    Dest := Pointer(DWORD(Dest) + head * VFilterDestUnitSize);
    XLoopCount := tail - head;
  end
  else
    XLoopCount := VPrm.XLoopCount;
  if XLoopCount <> 0 then
    _unsharpmask_vfilter(Dest, Src, @VPrm.WeightTableInfo[0], XLoopCount, VPrm.YLoopCount, SrcBpl, TmpBpl)
end;

end.
